/*
 *  
 *  Copyright (c) 2004 Steve Slaven, All Rights Reserved.
 *  
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *  MA 02111-1307 USA
 *  
*/
// This includes the map data handler, also handles caching and stuff like that
package net.hoopajoo.cheeselab3;
import java.io.*;
import java.util.*;
import jdbm.JDBMHashtable;
import jdbm.JDBMRecordManager;
import jdbm.JDBMEnumeration;

class CheeseMapData {
	private long last_map = 0,
		changed_map = 1,
		last_request_map = 0,
		request_map = 0;

	// These store the window for the last missed map request, so we can
	// re-run it later after the map has settled to make sure the
	// current view is up-to-date
	private Point3d request_start = new Point3d(), 
		request_end = new Point3d();

	private JDBMRecordManager db = null;
	private JDBMHashtable rooms = null;
	private JDBMHashtable coords = null; // persisten coords saving so that the map is usable quicker
	
	private HashMap players = new HashMap();
	private HashMap memrooms = new HashMap(); // faster memory access
	private HashMap memcoords = new HashMap(); // replaces the map items in cheesecanvas
	private HashMap lastrequest = new HashMap(); // last mtime we requested by roomid so
						// we don't make redundant requests
	private boolean use_memcache = false;

	// This is the highest seen mtime, not the last saved one
	private long highest_mtime = 0;

	private Vector remap_events = new Vector();

	// We save remap requests in a vector so we can clear out the map area
	// whenever we receive a map event, this makes it so that deleted rooms
	// will disappear
	class RemappedWindow {
		Point3d start = new Point3d(),
			end = new Point3d();
	}

	CheeseMapData() {
		CheeseLab.dmesg( 2, "Creating map data instance" );
		initcache();

		use_memcache = CheeseMap.prop_get_boolean( "mapdata_memcache" );
		CheeseLab.dmesg( 2, "Use memcache for map data: " + use_memcache );
	}

	// this is the last saved highest mtime
	public long get_highest_mtime() {
		return( highest_mtime );
	}
	
	public void initcache() {
		try {
			File cache = new File( CheeseMap.cachedir + "/tiles" );
			cache.mkdirs();
			
			db = new JDBMRecordManager( CheeseMap.cachedir + "/rooms.cache" );

			// Do we really need these? it's a cache after all...
			db.disableTransactions();

			rooms = db.getHashtable( "rooms" );
			coords = db.getHashtable( "coords" );			
		}catch( Exception e ) {
			CheeseLab.die( this, 1, "Unable to create room cache" );
		}

		// Clear memcache
		memrooms.clear();
		memcoords.clear();

		// get highest mtime from hash
		long max = Long.MAX_VALUE; // in case of exception do not refresh
		try {
			Long tmp_mtime = (Long)rooms.get( "highest_mtime" );
			if( tmp_mtime != null ) {
				CheeseLab.dmesg( 2, "Got last highest mtime: " + tmp_mtime );
				max = tmp_mtime.longValue();
			}else{
				max = 0;
			}
		}catch( Exception e ) {
			e.printStackTrace();
		}
		highest_mtime = max;

		if( CheeseMap.prop_get_boolean( "mapdata_memcache_preload" ) ) {
			// Preload rooms cache, kind of nasty
			CheeseLab.dmesg( 2, "Preloading room cache in to memory" );
			try {
				JDBMEnumeration keys = rooms.keys();
				while( keys.hasMoreElements() ) {
					Object key = keys.nextElement();
					memrooms.put( key, rooms.get( key ) );
				}
			}catch( Exception e ) {
				e.printStackTrace();
			}
		}
	}

	// Update a region of the map (most likely because we
	// are displaying it)
	public void remap( int sx, int sy, int sz, int ex, int ey, int ez ) {
		if( CheeseMap.prop_get_boolean( "sync_on_world" ) ) {
			// Do not use remap, using sync-on-world logic
			return;
		}


		if( last_map == changed_map ) {
			// TODO: we have not finished our last map update (which will inc last_map)
			// so discard this request (but note it so that a thread
			// can clean up this loose-end later)
			request_start.x = sx;	
			request_start.y = sy;	
			request_start.z = sz;	
			request_end.x = ex;	
			request_end.y = ey;	
			request_end.z = ez;	
			request_map++;
			CheeseLab.dmesg( 1, "Skipping overactive remap, queuing for later: " + request_map );
			return;
		}

		changed_map = last_map;

		RemappedWindow w = new RemappedWindow();
		w.start.x = sx;
		w.start.y = sy;
		w.start.z = sz;
		w.end.x = ex;
		w.end.y = ey;
		w.end.z = ez;

		remap_events.add( w );

		CheeseMap.send_tx( 1, "map",
			sx + "," + sy + "," + sz + " " +
			ex + "," + ey + "," + ez );
	}

	// This should be periodically called by a thread to make sure we
	// get caught up to the latest view if we had to skip out on some remap
	// requests
	public void catchup_remap() {
		if( request_map != last_request_map ) {
			CheeseLab.dmesg( 1, "Trying to catch up remap requests: " + last_request_map );
			last_request_map = request_map;
			remap( request_start.x, request_start.y, request_start.z,
				request_end.x, request_end.y, request_end.z );
		}
	}

	// This just queues room updates for stuff in our cache that is out
	// of date, this way we only get updates for what is in our
	// viewable area but we also can use old cached data in the mean time
	public void update_map( byte[] buffer ) {
		String s = new String( buffer );
		BufferedReader in = new BufferedReader( new StringReader( s ) );
		String line;

		// Get last remap req (if applicable)
		RemappedWindow w = (RemappedWindow)remap_events.get( 0 );
		HashMap clean_keys = new HashMap();
		if( w != null ) {
			remap_events.remove( 0 );

			// We make a vector of items to null-out based on all the keys
			// that could be set from this remap event.
			for( int x = w.start.x; x <= w.end.x; x++ ) {
				for( int y = w.start.y; y <= w.end.y; y++ ) {
					for( int z = w.start.z; z <= w.end.z; z++ ) {
						clean_keys.put( x + "," + y + "," + z, new Integer( 0 ) );
					}
				}
			}
		}

		try{
			while( (line = in.readLine()) != null ) {
				String[] values = line.split( "," );
				// Simply queue a request for the latest room data here,
				// if we already have something then it will be cached and
				// the renderer will do Do the Right Thing (tm) and render
				// what we have, and new room events will be generated as
				// updates come in and it'll repaint those
				CheeseRoomData room = get_room( Integer.parseInt( values[ 0 ] ), Long.parseLong( values[ 4 ] ) );
				
				// And poke what's to come in to our x/y/z array (since the final room
				// update will simply do it again, it's harmless but allows the renderer
				// to draw the ?? boxes where rooms will fill in to) and the room update
				// handler does the right thing to remove old reverse entries.  The only
				// drawback here is it's possible that a room may exist in 2 places on the
				// map at once until a room update is received
				//
				// TODO: It might be nice to automagically remove the old room someday though, maybe
				// if we had a common update_coords_cache() function that could be called here
				// or in the update_room subs that would be better?
				try {
					String coordkey = values[ 1 ] + "," + values[ 2 ] + "," + values[ 3 ];

					clean_keys.remove( coordkey );

					coords.put( coordkey, new Integer( values[ 0 ] ) );
					if( use_memcache ) {
						memcoords.put( coordkey, new Integer( values[ 0 ] ) );
					}
				}catch( Exception e ) {
					e.printStackTrace();
				}
			}

			// Any keys left in clean_keys are keys that need to be unset in the coords
			// and memcoords hashs
			Object[] keys = clean_keys.keySet().toArray();
			for( int c = 0; c < keys.length; c++ ) {
				coords.remove( keys[ c ] );
				if( use_memcache ) {
					memcoords.put( keys[ c ], new Integer( -1 ) );
				}
			}
		}catch( Exception e ){
			e.printStackTrace();
		}
		last_map++;
	}

	public void update_room( byte[] r ) {
		CheeseLab.dmesg( 2, "Updating room" );

		CheeseRoomData room = (CheeseRoomData)CheeseDecoder.decode_object( r );
		CheeseLab.dmesg( 2, "Decoded room: " + room.roomid + " - " +
			room.x + "," + room.y + "," + room.z + " - " + room.mtime );
	
		Integer roomid = new Integer( room.roomid );
		String coordkey = room.x + "," + room.y + "," + room.z;
			
		try {
			// Delete a reverse lookup if it's on file
			CheeseRoomData oldroom = (CheeseRoomData)rooms.get( roomid );
			if( oldroom != null ) {
				String oldcoordkey = oldroom.x + "," + oldroom.y + "," + oldroom.z;
				coords.remove( oldcoordkey );
			}
			
			// Save basic room data
			rooms.put( roomid, room );
			
			// Also add to reverse coords lookup
			coords.put( coordkey, roomid );
			
			if( room.mtime > highest_mtime ) {
				highest_mtime = room.mtime;
			}
			
			if( use_memcache ) {
				if( oldroom != null ) {
					String oldcoordkey = oldroom.x + "," + oldroom.y + "," + oldroom.z;
					// -1 means empty, null == cache miss
					memcoords.put( oldcoordkey, new Integer( -1 ) );
				}
				memrooms.put( roomid, room );
				memcoords.put( coordkey, roomid );
			}
		}catch( Exception e ) {
			CheeseLab.dmesg( 0, "Error saving object room: " + e );
			e.printStackTrace();
		}
	}

	// Returns the player that was updated
	public CheesePlayerData update_player( byte[] p ) {
		CheeseLab.dmesg( 2, "Updating player" );

		CheesePlayerData player = (CheesePlayerData)CheeseDecoder.decode_object( p );
		
		try {
			CheeseLab.dmesg( 3, "Saving player: " + player.name );
			//players.remove( player.name );
			players.put( player.name, player );
		}catch( Exception e ) {
			CheeseLab.dmesg( 0, "Error saving player object player: " + e );
			e.printStackTrace();
		}

		return( player );
	}

	public CheesePlayerData[] get_players() {
		Vector data = new Vector();
		try {
			//JDBMEnumeration values = players.values();
			Iterator values = players.values().iterator();
			while( values.hasNext() )
				data.add( values.next() );
		
			//while( values.hasMoreElements() )
			//	data.add( values.nextElement() );
		}catch( Exception e ) {
			CheeseLab.dmesg( 0, "Exception in get_players: " + e );
			e.printStackTrace();
		}
		return( (CheesePlayerData[])data.toArray( new CheesePlayerData[0] ) );
	}

	public CheesePlayerData get_player( String name ) {
		try {
			return( (CheesePlayerData)players.get( name ) );
		}catch( Exception e ) {
			return( null );
		}
	}

	// Get the room at coords x/y/z, may return NULL
	public CheeseRoomData get_room( int x, int y, int z ) {
		Integer roomid = null;
		String coordkey = x + "," + y + "," + z;

		if( use_memcache ) {
			roomid = (Integer)memcoords.get( coordkey );
			if( roomid == null ) {
				// Check disk cache too
				try {
					roomid = (Integer)coords.get( coordkey );
				}catch( Exception e ) {
					e.printStackTrace();
				}
				if( roomid != null ) {
					// update mem cache
					memcoords.put( coordkey, roomid );
				}else{
					memcoords.put( coordkey, new Integer( -1 ) );
				}
			}else{
				// A roomid of -1 in the memcoords means it was not a cache miss but
				// there is no room there
				// so we make it null here (otherwise missing rooms result in
				// cache misses and slow down rendering)
				if( roomid.intValue() == -1 )
					roomid = null;
			}
		}else{
			// Just check coords disk cache
			try {
				roomid = (Integer)coords.get( coordkey );
			}catch( Exception e ) {
				e.printStackTrace();
			}
		}

		if( roomid == null ) {
			return( null );
		}else{
			return( get_room( roomid.intValue() ) );
		}
	}

	// Give mtime as the servers latest mtime, if our room is
	// at least that new we will not even bother with a round-trip
	public CheeseRoomData get_room( int id ) {
		return( get_room( id, 1 ) );
	}
	
	public CheeseRoomData get_room( int id, long mtime ) {
		if( use_memcache )
			return( mem_get_room( id, mtime ) );

		// Skip memcache code
		return( disk_get_room( id, mtime ) );
	}

	private CheeseRoomData mem_get_room( int id, long mtime ) {
		// Get rooms from memcache first, if not there then 
		// ask the disk cache (which in turn asks the server
		// if it fails)
		CheeseRoomData room = (CheeseRoomData)memrooms.get( new Integer(id) );
		if( room != null )
			if( mtime > room.mtime )
				room = null;

		if( room == null ) {
			room = disk_get_room( id, mtime );
			memrooms.put( new Integer( id ), room );
		}
		return( room );
	}

	private CheeseRoomData disk_get_room( int id, long mtime ) {
		// These are not valid rooms
		if( id < 1 ) {
			CheeseLab.dmesg( 1, "Invalid roomid requested: " + id );
			CheeseLab.stacktrace();
			return( null );
		}

		try {
			CheeseRoomData room = (CheeseRoomData)rooms.get( new Integer(id) );

			if( room == null ) {
				// Send back a dummy room
				room = new CheeseRoomData();
				room.name = "Uncached Room (loading)";
				room.terrain = "?";
				room.overlays = new String[0];
				room.mtime = 0;
			}

			if( room.mtime < mtime ) {
				Long last_requesttime = (Long)lastrequest.get( new Integer( id ) );
				if( last_requesttime == null || last_requesttime.longValue() < mtime ) {
					CheeseLab.dmesg( 2, "Requesting room " + room.roomid + " (" + room.mtime + "<" + mtime + ")" );
					// Request it from the server, a refresh should
					// be issued when it comes in sooo :p
					// you ought to get a "filling in" effect while spamming about
					CheeseMap.send_tx( 1, "get", id + ":" + room.mtime );
					lastrequest.put( new Integer( id ), new Long( mtime ) );
				}
			}

			return( room );
		}catch( Exception e ) {
			return( null );
		}
	}
	
	public void save_max_mtime() {
		try {
			CheeseLab.dmesg( 2, "Saving highest_mtime: " + highest_mtime );
			rooms.put( "highest_mtime", new Long( highest_mtime ) );
		}catch( Exception e ) {
			e.printStackTrace();
		}
	}
}
