/*
 *  
 *  Copyright (c) 2004 Steve Slaven, All Rights Reserved.
 *  
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Incanvas., 59 Temple Place, Suite 330, Boston,
 *  MA 02111-1307 USA
 *  
*/
package net.hoopajoo.cheeselab3;
import java.util.*;
import java.awt.*;
import java.awt.geom.Rectangle2D;
import java.awt.image.*;

/**
 * 
 * Experimental threaded map renderer. Should prevent GUI freezes.
 * Based on Stevens's experimental cached rendering engine.
 * 
 * @author Antti Andreimann <Antti.Andreimann@mail.ee>
 */
class CheeseRendererThreaded extends CheeseRendererCached implements Runnable {
	Thread rendererThread = null;
	BufferedImage osc = null;
	BufferedImage oldosc = null;
	BufferedImage mapcache = null;
	Date last_render = null;
	
	int oldosc_x = 0, oldosc_y = 0, oldosc_z = 0;
	Object osclock;
	
	volatile int rapid_re_render_count = 0;
	volatile boolean stopThread = false,
					 waiting = false,
					 complete = false;
	volatile long needtorender = 0,
				  rendering = 0;

	CheeseRendererThreaded() {
		super();
		int prio = Thread.MIN_PRIORITY;
		
		osclock = new Object();
		
		CheeseLab.dmesg( 1, "Creating map painting thread" );
		
		rendererThread = new Thread( this, "MapPainterThread" );
		
		CheeseLab.dmesg( 2, "Setting thread priority to " + prio);
		try {
		    rendererThread.setPriority( prio );
		} catch (Exception e) {
		    CheeseLab.dmesg( 2, "Priority setting failed: " + e.getMessage() );
		}
		
		rendererThread.start();
	}

	public void prepare() {
		Dimension tile_size = canvas.get_tile_size();
		if( tile_width != tile_size.width ) {

		    synchronized(osclock) {
		        clearImage(oldosc);
		        clearImage(osc);
		    }
		}

		// if( render_grid != canvas.get_render_grid() ) {
		//	memorytiles.clear();
		//}

		// Let super update our properties
		super.prepare();		
	}
	
	public void run() {
	    CheeseLab.dmesg( 1, "Thread started" );
	    
	    while( !stopThread ) {
	        while( needtorender == rendering ) {
	            // Wait for repaint notification
	            synchronized(this) {
	                try {
	                    wait();
	                }
	                catch(InterruptedException e) {}
	            }
	        }

	        Date rstart = new Date();
	        complete = false;
	        
	        synchronized(osclock) {
	            if( osc == null && mapcache != null ) {
					CheeseLab.dmesg( 2, "Creating new buffered image for off-screen rendering" );
					osc = (BufferedImage)canvas.createImage( 
					        mapcache.getWidth(), mapcache.getHeight() );
	            }

		        if( last_render != null &&
		                oldosc_x == origin_x &&
		                oldosc_y == origin_y &&
		                oldosc_z == origin_z &&
		                rstart.getTime() - last_render.getTime() < 100)
		            rapid_re_render_count++;
		        else
		            rapid_re_render_count = 0;
	        }

	        if( rapid_re_render_count > 3 ) {
	            rapid_re_render_count = 0;
	            CheeseLab.dmesg( 2, "Rapid re-rendering with same origin detected, sleeping a bit to slow down");
	            try {
	                Thread.sleep( 500 );
	            } catch( InterruptedException e ) {}
	        }

	        if( drawMap() ) {
		        // Swap OSC
		        synchronized(osclock) {
		            BufferedImage tmp = osc;
		            osc = oldosc;
		            oldosc = tmp;
		            oldosc_x = origin_x;
		            oldosc_y = origin_y;
		            oldosc_z = origin_z;
		            mapcache.getGraphics().drawImage( oldosc, 0, 0, null );
		        }
		        
		        complete = true;

		        // TODO: Investigate the possibility of a race condition here (between check and lock)
			    if( waiting ) {
			        // Notify waiting thread that we are done
			        synchronized( this ) {
					    notifyAll();
			        }
			    }
				else {
				    // Or request a repaint if nobody is waiting anymore
					canvas.changed();
				}
	        }
	        else {
	            CheeseLab.dmesg( 2, "Rendering aborted due to new redraw!");
	        }
	        
			Date rend = new Date();
			long rdiff = rend.getTime() - rstart.getTime();
			double secs = (double)rdiff / 1000;
			CheeseLab.dmesg( 1, "Render time: " + rdiff + "ms (" + secs + " secs)" );
			
			/* NB! This line must be here and executed even if rendering
			 * was aborted in order to detect rapid re-renders
			 */
			last_render = rend;
	    }
	}
	
	protected void clearImage( BufferedImage o ) {
	    if( o == null )
	        return; // Nothing to do
	    
	    clearImage(o, 0, 0, o.getWidth(), o.getHeight() );
	}
	
	protected void clearImage( BufferedImage o, int x, int y, int w, int h ) {
	    if( o == null )
	        return; // Nothing to do

	    o.getGraphics().setColor( Color.black );
	    o.getGraphics().fillRect( x, y, w, h );
	}

	public void draw_map( BufferedImage buffer ) {
	    Graphics g = buffer.getGraphics();
	    
		// Make sure the map matches the size of it
		synchronized( osclock ) {
		    mapcache = buffer;
		    
		    if( osc != null ) {
		        if( osc.getWidth() != buffer.getWidth() ) {
		            osc.flush();
		            osc = null;
		        }else if( osc.getHeight() != buffer.getHeight() ) {
		            osc.flush();
		            osc = null;
		        }
		    }
		}

		// Notify rendering thread that it should start it's work
        needtorender++;
	    synchronized( this ) {
	        notifyAll();
	    }
	    
	    /* Wait for rendering to complete but not for more than 20ms.
	     * The idea behind this arrangement is that user will probably not
	     * notice a GUI "freeze" that lasts for less than 20ms and this way
	     * we save some repaints for fast rendering cases and thus
	     * reduce CPU load.
	     */
	    waiting = true;
	    synchronized( this ) {
	        try {
	            wait( 20 );
	        } catch( InterruptedException e ) {}
	    }
	    waiting = false;
	    
	    if( complete )
	        return;

	    /* If rendering didn't complete in time, we take an old osc 
	     * and draw it to new position (if possible).
	     */
		synchronized( osclock ) {
		    if( oldosc != null && origin_z == oldosc_z ) {
		        int offset_x = 0, offset_y = 0;
		        
		        Point p = canvas.global_to_local( oldosc_x, oldosc_y, oldosc_z );
		        offset_x = p.x;
		        offset_y = p.y;
		            
		        CheeseLab.dmesg( 3, "Drawing old screen to offset: " + offset_x + "," + offset_y);
		    	g.drawImage( oldosc, offset_x, offset_y, null );
		    	
		    	// Fill the rest with black
		    	g.setColor( Color.black );
		    	if( offset_x > 0)
		    	    g.fillRect( 0, 0, offset_x, buffer.getHeight() );
		    	else
		    	    g.fillRect( buffer.getWidth() + offset_x, 0, -offset_x, buffer.getHeight() );
		    	if( offset_y > 0)
		    	    g.fillRect( 0, 0, buffer.getWidth(), offset_y );
		    	else
		    	    g.fillRect( 0, buffer.getHeight() + offset_y, buffer.getWidth(), -offset_y );
		    }
		    else {
		        clearImage( buffer );
		    }
		}
		
		// Draw visual indication
		String str = "Loading ... ";
        FontMetrics metrics = g.getFontMetrics();
        Rectangle2D bounds = metrics.getStringBounds( str, g );
        g.setColor( Color.LIGHT_GRAY );
        g.drawString( str,
        (int)(buffer.getWidth() - bounds.getWidth()) + 1,
        (int)(buffer.getHeight() - bounds.getHeight()) + 1 );	    
	}
	
	// This function draws just the map, not player tiles or HUD data
	// which may change without the map data moving
	protected boolean drawMap() {
	    rendering = needtorender;
	    int cols, rows, width, height;
	    Graphics2D sg;
	    
	    synchronized(osclock) {
	        if( osc == null )
	            return false; // Nothing to do (yet)

	        cols = osc.getWidth() / tile_width;
	        rows = osc.getHeight() / tile_height;
	        sg = (Graphics2D)osc.getGraphics();
	        width = osc.getWidth();
	        height = osc.getHeight();
	    }
	    
	    CheeseLab.dmesg( 3, "Drawing!" );
	    
		// We mark the sync now so that if there are changed while we
		// are rendering, we will get them
		int rend_x = origin_x,
			rend_y = origin_y,
			rend_z = origin_z;

		CheeseLab.dmesg( 3, "Size: " + rows + "," + cols );
		CheeseLab.dmesg( 3, "Origin: " + rend_x + "," + rend_y + "," + rend_z );
			
		// Figure our bounding box
		int start_x = rend_x;
		int start_y = rend_y;
		int start_z = rend_z;

		int end_x = start_x + cols;
		int end_y = start_y + rows;
		int end_z = rend_z + 1;

		warplines.clear();
			
		// Make a slightly larger font
		sg.setFont( new Font( null, Font.BOLD, 12 ) );
			
		sg.setColor( new Color( 50, 50, 50 ) );
		sg.fillRect( 0, 0, width, height );
			
		Vector textnodes = new Vector();
		int x, y, z, local_x, local_y;

		for( z = 1; z >= 0; z-- ) {
			for( x = 0; x <= cols; x++ ) {
				local_x = x * tile_width;
				for( y = 0; y <= rows; y++ ) {
				    
				    // If needtorender changes, it means we are lagging behind
				    if (rendering != needtorender)
				        return false;
				    
					local_y = y * tile_height;
					int global_x = start_x + x;
					int global_y = start_y + y;
					int global_z = start_z + z;

					CheeseRoomData room = CheeseMap.map.get_room( global_x, global_y, global_z );
					if( room != null ) {
						RenderCacheItem thisroom = get_cached_tile( room );
						sg.drawImage( thisroom.image, local_x, local_y, null );

						warplines.addAll( thisroom.warplines );

						if( global_z == rend_z ) {
							// render notes text
							if( room.notes != null && (! room.notes.equals( "" ) ) ) {
								if( room.notes.length() < canvas.get_render_notes_length() && canvas.get_render_notes() ) {
									textnodes.add( room );
								}else{
									if( ! thisroom.drew_an_overlay ) {
										// Render a note tile if notes are here
										CheeseLab.dmesg( 3, "Doing default notes overlay" );
										sg.drawImage( get_image( "overlay_notes" ), local_x, local_y, null );
									}
								}
							}
							
						}else{
							sg.drawImage( get_image( "overlay_blacken" ), local_x, local_y, null );
						}
						
						// Selected
						if( canvas.room_is_selected( room.roomid ) ) {
							CheeseLab.dmesg( 3, "Drawing selected" );
							sg.drawImage( get_image( "selected" ), local_x, local_y, null );   
						}
						
						if( room.roomid == CheeseMap.edit_roomid )
							sg.drawImage( get_image( "infobox" ), local_x, local_y, null );

					}
				}
			}
		}

		// warp lines
		sg.setColor( new Color( 255, 0, 0 ) );
		for( int i = 0; i < warplines.size(); i++ ) {
		    // If needtorender changes, it means we are lagging behind
		    if (rendering != needtorender)
		        return false;
		   
			int[] points = (int[])warplines.get( i );
			// only on this zplane
			if( points[ 2 ] == rend_z ) {
				Point pstart = canvas.global_to_local_reference( points[ 0 ], points[ 1 ], points[ 2 ],
						rend_x, rend_y, rend_z );
				Point pend = canvas.global_to_local_reference( points[ 3 ], points[ 4 ], points[ 5 ],
						rend_x, rend_y, rend_z );

			
				sg.drawLine( pstart.x + points [ 6 ], pstart.y + points[ 7 ],
					pend.x + tile_width / 2, pend.y + tile_height / 2 );
			}
		}
		
		// Notes
		for( int i = 0; i < textnodes.size(); i++ ) {
		    // If needtorender changes, it means we are lagging behind
		    if (rendering != needtorender)
		        return false;
		   
			CheeseRoomData room =(CheeseRoomData)textnodes.get( i );
			Point p = canvas.global_to_local_reference( room.x, room.y, room.z, rend_x, rend_y, rend_z );
			local_x = p.x;
			local_y = p.y;
			
			int offset_x = 4;
			int offset_y = tile_height - 4;
			
			if( room.x % 2 == 0 ) {
				offset_y = tile_height / 2;
			}
			
			sg.setColor( Color.BLACK );
			sg.drawString( room.notes, 
				local_x + offset_x, local_y + offset_y );
			sg.setColor( Color.LIGHT_GRAY );
			sg.drawString( room.notes, 
				local_x + offset_x + 1, local_y + offset_y + 1 );
		}

		sg.dispose();
		
		return true;
	}
}
